import re


def replace_all(text):
    # Replace all non word characters
    text = text.replace(" ", "")
    return re.sub("\\W|_", "-", text)


def remove_multiple_hyphens(string):
    # make sure no double hyphens, AWS won't allow
    cleaned = re.sub(r"-+", "-", string)
    return cleaned


def clean_name_string(raw_string, truncate_len=40):
    cleaned = replace_all(raw_string).lower()
    cleaned = cleaned.strip("-")
    cleaned = remove_multiple_hyphens(cleaned)

    if len(cleaned) > truncate_len:
        truncated = cleaned[:truncate_len].strip("-")
    else:
        truncated = cleaned

    return truncated



from typing import Optional
from google.cloud import secretmanager

from cnc.models import EnvironmentCollection
from typing import Literal

from cnc.logger import get_logger

log = get_logger(__name__)


class GCPEnvironmentCollection(EnvironmentCollection):
    provider: Literal["gcp"]
    allow_net_admin: Optional[bool] = False

    def get_secret_value(self, secret_id, version_id="latest"):
        """
        Access a secret version in Secret Manager.

        Args:
        secret_id: ID of the secret you want to access.
        version_id: Version of the secret; defaults to "latest".

        Returns:
        The secret value as a string.
        """
        # Create the Secret Manager client.
        client = secretmanager.SecretManagerServiceClient()

        # Build the resource name of the secret version.
        name = f"projects/{self.account_id}/secrets/{secret_id}/versions/{version_id}"

        # Access the secret version.
        response = client.access_secret_version(request={"name": name})

        # Return the decoded payload.
        return response.payload.data.decode("UTF-8")

    @property
    def bastion_instance_type(self):
        if self.region == "southamerica-west1":
            return "e2-micro"
        else:
            return "f1-micro"

    @property
    def has_service_domains(self):
        if self.application.flavor == "run-lite":
            return True
        return False

    def cli_info_for_environment(self, environment):
        return {
            "name": environment.name,
            "load_balancer_ip": self.get_terraform_output("load_balancer_ip"),
            "domains": environment.domains,
        }

    def preconfigure_infrastructure(self, should_cleanup=True):
        config = self.provision_stage_manager
        config.make_ready_for_use(should_cleanup=should_cleanup)

        results = []
        steps = []

        total_plan_changes = config.plan(save=True, plan_filename="total_1")
        if not total_plan_changes:
            log.debug(
                "Did not get any total plan changes! Calling apply and logging..."
            )
            try:
                _ret = config.apply()
            except Exception as e:
                log.debug(f"Error in calling apply... {e}")
                _ret = {"@message": f"Error: {e}"}
            results.append(_ret)

            log.debug(f"Exiting infra configure step for {self}")
            return {"ok": False, "results": results, "steps": steps}

        # eventually we should wrap this with our own lock, then it would be safe to force unlock and proceed...
        if total_plan_changes["status"] == "locked":
            steps.append(total_plan_changes)
            return {"ok": False, "results": results, "steps": steps}

        plan_json = config.plan_json(plan_filename="total_1") or {}
        total_plan_changes["resource_changes"] = {
            "create": [],
            "update": [],
            "delete": [],
        }
        for rc in plan_json.get("resource_changes", []):
            actions = rc["change"]["actions"]
            resource_data = {key: rc[key] for key in rc.keys() if key not in ["change"]}
            resource_data["actions"] = actions
            if "create" in actions:
                total_plan_changes["resource_changes"]["create"].append(resource_data)
            elif "delete" in actions:
                total_plan_changes["resource_changes"]["delete"].append(resource_data)
            elif "update" in actions:
                total_plan_changes["resource_changes"]["update"].append(resource_data)

        steps.append(total_plan_changes)

        return {"ok": True, "results": results, "steps": steps}

    def configure_infrastructure(self, should_cleanup=True):
        can_configure = True
        config = self.provision_stage_manager

        results = []
        steps = []

        if not config.saved_plan_exists(plan_filename="total_1"):
            res = self.preconfigure(should_cleanup=should_cleanup)
            if not res["ok"]:
                return res

        total_plan_json = config.plan_json(plan_filename="total_1") or {}
        all_changes = total_plan_json.get("resource_changes", [])

        if self.contains_db_replacement_changes(plan_json=total_plan_json):
            self.process_db_replacement_alert()
            return {"ok": False, "results": results, "steps": steps}

        _ok = False
        ret = config.apply(use_saved=True, plan_filename="total_1")
        if ret and ret.get("error"):
            results.append(ret)
            if ret["error"] == "enable_billing":
                log.info("Cannot proceed with infra config, need billing account!")
                can_configure = False

        seen_url_addresses = []
        if can_configure:
            for change in all_changes:
                if (change["type"] == "google_compute_backend_service") and (
                    ("delete" in change["change"]["actions"])
                    or ("create" in change["change"]["actions"])
                ):
                    # log.debug(
                    #     f"Detected backend service/url map change for {self}, targeting url map and https proxy before continuting"
                    # )

                    if change["address"] in seen_url_addresses:
                        log.debug(f"Skipping {change}")
                        continue
                    else:
                        seen_url_addresses.append(change["address"])

                    # log.debug(f"Processing {change}")

                    url_map_changes = config.plan(
                        save=True, target=change["address"], plan_filename="urlmap"
                    )
                    url_map_changes["change"] = change

                    # log.info(f"Going to apply URL map changes: {url_map_changes}")
                    config.apply(use_saved=True, plan_filename="urlmap")

                if (change["type"] == "google_compute_target_https_proxy") and (
                    change["change"]["actions"] != ["no-op"]
                ):

                    https_proxy_changes = config.plan(
                        save=True,
                        target=f"google_compute_target_https_proxy.{self.instance_name}",
                        plan_filename="httpsproxy",
                    )
                    https_proxy_changes["change"] = change

                    # log.info(f"Going to apply HTTPS proxy changes: {https_proxy_changes}")
                    config.apply(use_saved=True, plan_filename="httpsproxy")

            services_to_cleanup = []
            for change in total_plan_json["resource_changes"]:
                if (
                    (change["type"] == "google_compute_backend_service")
                    and ("delete" in change["change"]["actions"])
                    and (len(change["change"]["actions"]) == 1)
                ) or (
                    (change["type"] == "google_compute_backend_bucket")
                    and ("delete" in change["change"]["actions"])
                    and (len(change["change"]["actions"]) == 1)
                ):
                    # log.debug(
                    #     f"Got stale backend that will need to be cleaned up: {change['address']}"
                    # )
                    services_to_cleanup.append(change["address"])

            total_plan_changes = config.plan(save=True, plan_filename="total")
            total_plan_json = config.plan_json(plan_filename="total") or {}
            if self.contains_db_replacement_changes(plan_json=total_plan_json):
                self.process_db_replacement_alert()
                return {"ok": False, "results": results, "steps": steps}

            log.info(f"Going to apply total plan changes: {total_plan_changes}")
            config.apply(use_saved=True, plan_filename="total")

            addresses_to_clean = []
            for i, service_address in enumerate(services_to_cleanup):
                if service_address in addresses_to_clean:
                    continue
                else:
                    addresses_to_clean.append(service_address)

                # log.debug(f"Cleaning up stale service for {service_address}")
                config.plan(
                    save=True,
                    target=service_address,
                    plan_filename=f"serviceaddress-{i}",
                )
                config.apply(use_saved=True, plan_filename=f"serviceaddress-{i}")
                # log.debug(f"All set cleaning up...")

            # One last apply to properly report success/fail
            _ret = config.apply()

            _ok = True
            if _ret.get("@message", "").startswith("Error: "):
                _ok = False

            results.append(_ret)

            # outputs = config.output()

            config.cleanup()

        return {"ok": _ok, "results": results, "steps": steps}

    def contains_db_replacement_changes(self, plan_json):
        db_resource_types = ["google_sql_database_instance", "google_sql_database"]
        try:
            all_changes = plan_json.get("resource_changes", [])
            for change in all_changes:
                if (
                    change["type"] in db_resource_types
                    and ("delete" in change["change"]["actions"])
                    and ("create" in change["change"]["actions"])
                ):
                    return True
        except Exception as e:
            log.debug(f"Error while checking for db replacement changes: {e}")

        return False

    def process_db_replacement_alert(self):
        """
        Adds an error to configuration results and sends an
        internal slack notification
        """
        self.results.append(
            {
                "@message": "Error: Unsafe operation - Planned database resource replacement detected."
            }
        )

    @property
    def service_identity_email(self):
        return (
            f"{self.cloud_resource_namespace}@{self.account_id}.iam.gserviceaccount.com"
        )



from unittest.mock import patch

from .base_test_class import CNCBaseTestCase
from cnc.models import Application
from cnc.constants import EnvironmentVariableTypes

from cnc.logger import get_logger

log = get_logger(__name__)


class ServiceProviderLinksTestCase(CNCBaseTestCase):
    fixture_name = "backend-1-service"

    def setUp(self):
        super().setUp()
        app = Application.from_environments_yml("environments.yml")
        self.collection = app.collections[0]
        self.environment = self.collection.environments[0]
        self.assertEqual(self.environment.name, "main")
        self.service = self.environment.services[0]

    def test_service_provider_links(self):
        self.assertEqual(
            self.service.provider_links[0]["url"],
            "https://console.cloud.google.com/run/detail/us-east1/c18d7ebb6d-my-ba-eview-main-app/revisions?project=foo-bar-123",
        )


class ServiceDomainsTestCase(CNCBaseTestCase):
    fixture_name = "backend-1-service"

    def test_service_domain_present(self):
        app = Application.from_environments_yml("environments_gcp_run_lite.yml")
        self.assertEqual(app.flavor, "run-lite")
        collection = app.collections[0]
        self.assertEqual(collection.has_service_domains, True)
        environment = collection.environments[0]
        service = environment.services[0]

        with patch(
            "cnc.models.providers.google.environment_collection.GCPEnvironmentCollection.get_terraform_output",
            return_value="https://foo-bar-123.run.app",
        ):
            self.assertEqual(service.domain, "https://foo-bar-123.run.app")
            self.assertEqual(
                environment.domains,
                [
                    {
                        "service_name": service.name,
                        "domain": "https://foo-bar-123.run.app",
                    }
                ],
            )

    def test_environment_domain_present(self):
        app = Application.from_environments_yml("environments.yml")
        collection = app.collections[0]
        self.assertEqual(collection.has_service_domains, False)
        environment = collection.environments[0]
        service = environment.services[0]

        self.assertEqual(service.domain, None)
        self.assertEqual(len(environment.domains), 1)


class ServiceEnvironmentItemsTestCase(CNCBaseTestCase):
    fixture_name = "backend-1-service-1-db-variables-aliases"

    def setUp(self):
        super().setUp()
        app = Application.from_environments_yml("environments.yml")
        self.collection = app.collections[0]
        self.environment = self.collection.environments[0]
        self.assertEqual(self.environment.name, "main")
        self.service = self.environment.services[0]
        self.assertEqual(self.service.name, "app")

    def test_all_items(self):
        # outputs - 2
        # secrets - 2
        # standard vars - 2
        total_env_items = 6
        total_env_items += len(self.service.config.managed_environment_variables)
        total_env_items += len(self.service.config.managed_environment_secrets)

        self.assertEqual(len(self.service.environment_items), total_env_items)

    def test_environment_secrets(self):
        num_secrets = 2
        num_secrets += len(self.service.config.managed_environment_secrets)
        self.assertEqual(len(self.service.environment_secrets), num_secrets)

        service_secrets = self.service.filtered_environment_items(
            variable_type=EnvironmentVariableTypes.VARIABLE_TYPE_SECRET
        )
        self.assertEqual(service_secrets[0].name, "foo-secret")
        self.assertEqual(service_secrets[1].name, "foo-secret-alias")
        self.assertEqual(service_secrets[0].secret_id, service_secrets[1].secret_id)

    def test_environment_outputs(self):
        self.assertEqual(len(self.service.environment_outputs), 2)

        service_secrets = self.service.environment_outputs
        self.assertEqual(service_secrets[0].name, "foo-output")
        self.assertEqual(service_secrets[1].name, "foo-output-alias")
        self.assertEqual(service_secrets[0].output_name, service_secrets[1].output_name)

    def test_environment_standard_variables(self):
        service_vars = self.service.filtered_environment_items(
            variable_type=EnvironmentVariableTypes.VARIABLE_TYPE_STANDARD
        )
        self.assertEqual(len(service_vars), 2)

        self.assertEqual(service_vars[0].name, "foo-standard")
        self.assertEqual(service_vars[1].name, "foo-standard-alias")
        self.assertEqual(service_vars[0].value, service_vars[1].value)

    def test_environment_variables(self):
        # num default managed vars for all envs - 3
        # num standard vars - 2
        num_service_env_vars = 5

        for svc in self.environment.config.services:
            num_service_env_vars += len(svc.settings.managed_environment_variables)

        service_vars = self.service.environment_variables
        self.assertEqual(len(service_vars), 15)



from typing import Optional, Union
from pydantic import Field

from cnc.models.base_model import BaseModel
from cnc.logger import get_logger

log = get_logger(__name__)


class DeployResourceLimits(BaseModel):
    raw_memory: Optional[str] = Field(default="512Mi", alias="memory")
    raw_cpu: Optional[Union[int, float]] = Field(default=1, alias="cpus")



from .base_test_class import CNCBaseTestCase
from cnc.models import Application, DeployStageManager


from cnc.logger import get_logger

log = get_logger(__name__)


class DeployStageTestBase(CNCBaseTestCase):
    fixture_name = "backend-1-service-1-db"
    environment_name = "main"

    def setUp(self):
        super().setUp()
        app = Application.from_environments_yml(self.env_data_filepath)
        collection = app.collections[0]
        collection._infra_outputs_cache = {}

        self.environment = collection.environment_by_name(self.environment_name)
        self.assertEqual(self.environment_name, self.environment.name)

        self.deployer = DeployStageManager(self.environment)

        self.deployer.cleanup()
        self.deployer.setup()
        self.deployer.render_scripts()

    def tearDown(self):
        super().tearDown()
        if hasattr(self, "deployer"):
            self.deployer.cleanup()

    def parse(self, script_name="build.sh"):
        with open(f"{self.deployer.rendered_files_path}/{script_name}", "r") as file:
            return file.read()


class AWSDeployStageTestBase(DeployStageTestBase):
    env_data_filepath = "environments_aws_ecs.yml"

    def setUp(self):
        super().setUp()
        self.assertTrue(self.environment.application.provider_is_aws)


class DeployStageCustomTestBase(DeployStageTestBase):
    fixture_name = "backend-1-service-custom"

    def setUp(self):
        super().setUp()

    def test_custom_template(self):
        build = self.parse("deploy-app.sh")
        self.assertIn("custom_template", build)


class AWSDeployStageSmokeTest(AWSDeployStageTestBase):
    def test_deploy_functions(self):
        build = self.parse("deploy-app.sh")
        self.assertNotIn("deploy_app_scheduled_tasks", build)
        self.assertNotIn("deploy_app_workers_to_ecs", build)
        self.assertIn(
            (
                f"source { self.deployer.rendered_files_path }/deploy-"
                "app-functions.sh"
            ),
            build,
        )
        self.assertIn("deploy_app_to_ecs", build)
        self.assertIn("check_app_ecs_service_deploy_status", build)

        build_functions = self.parse("deploy-app-functions.sh")
        self.assertNotIn("deploy_app_scheduled_tasks", build_functions)
        self.assertNotIn("deploy_app_workers_to_ecs", build_functions)
        self.assertIn("deploy_app_to_ecs", build_functions)
        self.assertIn("check_app_ecs_service_deploy_status", build_functions)

        for service in self.environment.web_services:
            web_task = self.parse(f"ecs-web-{service.name}.json")
            self.assertIn(service.instance_name, web_task)



from typing import Optional

from .base_model import BaseModel


class Stage(BaseModel):
    name: str
    provider_token: Optional[str] = ""
    callback_url: Optional[str] = ""

    def __getattribute__(self, item):
        # Override to catch "is_<name>" pattern
        if item.startswith("is_"):
            name_to_check = item[3:]

            if name_to_check not in ["build", "deploy"]:
                # preserve default attributes of BaseModel if not a valid cnc stage name
                return super().__getattribute__(item)

            # Compare with the instance's name property
            if name_to_check == super().__getattribute__("name"):
                return True
            return False
        else:
            return super().__getattribute__(item)



from typing import Literal
from cnc.models.config.resource import CacheResourceSettings
from cnc.logger import get_logger

log = get_logger(__name__)


class GCPCacheResourceSettings(CacheResourceSettings):
    provider: Literal["gcp"]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def provider_version(self):
        _version = float(self.version)
        if _version == 7.2:
            return "REDIS_7_2"

        major_version = int(_version)
        if major_version == 6:
            return "REDIS_6_X"

        return "REDIS_{{ service.settings.version }}_0"

    @property
    def toolbox_ssh_port_mapping(self):
        return f"{self.local_port}:{self.redis_ip}:{self.redis_port}"
